from fnmatch import fnmatch
try:
    import ConfigParser as configparser
except ImportError:
    import configparser
from logsetup import UserLogger
from errors import EXIT_CODE_CONFIG_FILE_NOT_FOUND

log = UserLogger(__name__)

class dummy_context_mgr():

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        return False

def load_config(arguments):
    config = configparser.ConfigParser(allow_no_value=True)
    configpath = arguments['--configfile']

    try:
        with open(configpath) as f:
            config.readfp(f)
    except IOError:
        log.uabort('Config file path not found. ({})', configpath,
                   exit_code=EXIT_CODE_CONFIG_FILE_NOT_FOUND)

    config_dict = {}
    for section in config.sections():
        config_dict[section] = dict((key, True if value is None else value)
                                    for key, value in config.items(section))
    return config_dict


def update_arguments_to_config(arguments, config):
    def key_set(section, key, value):
        s = config.get(section, dict())
        config[section] = s
        s[key] = value
    lookup = {
        "--hostname": ('infinibox', 'host'),
        "--retain": ('global', 'snaps_to_keep'),
        "--basename": ('global', 'snap_basename'),
        "--debug": ('logging', 'debug'),
        "--logfile": ('logging', 'logpath'),
        "--norotation": ('logging', 'rotation'),
        "--password": ('infinibox', 'password'),
        "--quiet": ('logging', 'quiet'),
        "--syslog": ('logging', 'syslog'),
        "--user": ('infinibox', 'username'),
        "--volumes": ('infinibox', 'volumes'),
        "--purge": ('infinibox', 'purge'),
        "--dryrun": ('global', 'dryrun'),

    }
    for arg_key, (section, key) in lookup.items():
        if arg_key in arguments and arguments[arg_key] != None and arguments[arg_key] != False:
            key_set(section, key, arguments[arg_key])


def validate_config(config):
    from cerberus import Validator

    class MyValidator(Validator):

        def _validate_bool(self, param, field, value):
            if str(value).lower() not in ['true', 'false', 'yes', 'no', 'on', 'off', '1', '0']:
                self._error(field, "Not in bool list")
            else:
                if str(value).lower() in ['true', 'yes', 'on', '1']:
                    self.document[param][field] = True
                else:
                    self.document[param][field] = False

        def _validate_retention(self, param, field, value):
            if not value.isdigit():
                self._error(field, "Not a digit list")
            else:
                retent = int(value)
                self.document[param][field] = retent
                if retent > 254 or retent < 1:
                    self._error(field, "Retention must be between 1 and 254")

        def _validate_parallel(self, param, field, value):
            if not value.isdigit():
                self._error(field, "Not a digit list")
            else:
                parallel = int(value)
                self.document[param][field] = parallel
                if parallel > 20 or parallel < 1:
                    self._error(field, "Parallel must be between 1 and 20")

    schema = {
        'global': {'type': 'dict', 'required': True, 'schema': {
            'snap_basename': {'type': 'string', 'required': True},
            'dryrun': {'bool': 'global'},
            'use_volid': {'bool': 'global'},
            'sa_internal': {'bool': 'global'},
            'single_instance': {'bool': 'global'},
            'snaps_to_keep': {'retention': 'global'}}},
        'infinibox': {'type': 'dict', 'required': True, 'schema': {
                      'host': {'type': 'string', 'dependencies': ['password', 'username'], 'required': True},
                      'password': {'type': 'string', 'dependencies': ['host', 'username'], 'required': True},
                      'username': {'type': 'string', 'dependencies': ['password', 'host'], 'required': True},
                      'purge': {'bool': 'infinibox', 'dependencies': ['volumes']},
                      'volumes': {'type': 'string', 'dependencies': ['host'], 'required': False, 'empty': False}}},
        'logging': {'type': 'dict', 'schema': {
            'logpath': {'type': 'string'},
            'emailuser': {'type': 'string'},
            'emailpassword': {'type': 'string'},
            'emailsender': {'type': 'string'},
            'emailrcpt': {'type': 'string'},
            'emailserver': {'type': 'string'},
            'emailport': {'type': 'string'},
            'quiet': {'bool': 'logging'},
            'rotation': {'bool': 'logging'},
            'debug': {'bool': 'logging'},
            'email': {'bool': 'logging'},
            'suppress': {'bool': 'logging'},
            'syslog': {'bool': 'logging'}}},
        'vcenter': {'type': 'dict', 'dependencies': ['vms', 'noquiesce', 'nomemory', 'datastores'], 'required': False, 'schema': {
            'host': {'type': 'string', 'dependencies': ['password', 'username'], 'required': False},
            'password': {'type': 'string', 'dependencies': ['host', 'username'], 'required': False},
            'parallel': {'parallel': 'vcenter'},
            'username': {'type': 'string', 'dependencies': ['host', 'password'], 'required': False}}},
        'datastores': {'type': 'dict', 'required': False},
        'nomemory': {'type': 'dict', 'required': False},
        'noquiesce': {'type': 'dict', 'required': False},
        'vms': {'type': 'dict', 'required': False},
        'volumes': {'type': 'dict', 'required': False},
        'fs': {'type': 'dict', 'required': False},
    }

    v = MyValidator(schema)
    result = v.validate(config)

    if 'infinibox' not in config:
        result = False
    else:
        if 'volumes' not in config['infinibox'] and 'vcenter' not in config and 'volumes' not in config and 'fs' not in config:
            result = False
            v._error(
                'volumes', 'Either [vmware] section or [volumes] or [fs] or --volumes should be configured.')

        elif 'volumes' in config['infinibox'] and 'vcenter' in config:
            result = False
            v._error(
                'volumes', '[vmware] section and --volumes can\'t be configured simultaneously.')

        elif 'volumes' in config and 'vcenter' in config:
            result = False
            v._error(
                'volumes', '[vmware] section and [volumes] can\'t be configured simultaneously.')

        elif 'fs' in config and 'vcenter' in config:
            result = False
            v._error(
                'fs', '[vmware] section and [fs] can\'t be configured simultaneously.')

        elif 'volumes' in config and 'volumes' in config['infinibox']:
            result = False
            v._error(
                'volumes', '--volumes (or infinibox[volumes]) section and [volumes] can\'t be configured simultaneously.')
        elif 'fs' in config and 'volumes' in config['infinibox']:
            result = False
            v._error(
                'volumes', '--volumes (or infinibox[volumes]) section and [fs] can\'t be configured simultaneously.')

        if 'global' not in config:
            result = False
            v._error(
                'snap_basename', 'snap_basename (--basename or -b) and snaps_to_keep (--retain or -r)/--purge is required.')
        else:
            if 'snaps_to_keep' in config['global'] and 'purge' in config['infinibox']:
                if config['infinibox']['purge']:
                    result = False
                    v._error(
                        'purge', 'Either --purge or --snaps_to_keep should be specified, not both.')
            elif 'snaps_to_keep' not in config['global'] and 'purge' not in config['infinibox']:
                result = False
                v._error(
                    'snaps_to_keep', 'Either --purge or --snaps_to_keep should be specified')

            if 'purge' in config['infinibox']:
                if config['infinibox']['purge']:
                    config['global']['snaps_to_keep'] = 1
    if 'logging' in config and 'email' in config['logging'] and config['logging']['email'] == True:
        if 'emailsender' not in config['logging'] \
                or 'emailrcpt' not in config['logging'] \
                or 'emailserver' not in config['logging']:
            v._error(
                'logging', 'emailsender,emailrcpt,emailserver are mandatory when email is enabled')
            result = False
        if 'emailuser' in config['logging'] or 'emailpassword' in config['logging']:
            if 'emailuser' not in config['logging'] or 'emailpassword' not in config['logging']:
                v._error(
                    'logging', 'emailuser and emailpassword shoud both be present or absent')
                result = False

    return result, v


def read_datastore_patterns_from_config(config):
    try:
        return config['datastores']
    except KeyError:
        return []


def read_vm_patterns_from_config(config):
    try:
        return config['vms']
    except KeyError:
        return []


def read_noquiesce_patterns_from_config(config):
    try:
        return config['noquiesce']
    except KeyError:
        return []


def read_nomemory_patterns_from_config(config):
    try:
        return config['nomemory']
    except KeyError:
        return []


def read_volumes_patterns_from_config(config):
    try:
        return config['volumes']
    except KeyError:
        return []


def read_filesystems_patterns_from_config(config):
    try:
        return config['fs']
    except KeyError:
        return []


def iter_filtered_datastores(datastores, pattern):
    for datastore in datastores:
        if fnmatch(datastore.name.lower(), pattern.lower()):
            yield datastore


def get_datastores_to_snapshot(datastores, patterns):
    def get_filtered_datastores(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_datastores(datastores, pattern))
        return result

    includes = get_filtered_datastores(lambda p: p and p[0] != '!')
    excludes = get_filtered_datastores(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_volumes_to_snapshot(config, volumes):
    patterns = read_volumes_patterns_from_config(config)
    return get_volumes_to_snapshot(volumes, patterns)


def iter_filtered_volumes(volumes, pattern):
    for volume in volumes:
        if fnmatch(volume.lower(), pattern.lower()):
            yield volume


def get_volumes_to_snapshot(volumes, patterns):
    def get_filtered_volumes(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_volumes(volumes, pattern))
        return result

    includes = get_filtered_volumes(lambda p: p and p[0] != '!')
    excludes = get_filtered_volumes(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_filesystems_to_snapshot(config, filesystems):
    patterns = read_filesystems_patterns_from_config(config)
    return get_filesystems_to_snapshot(filesystems, patterns)


def iter_filtered_filesystems(filesystems, pattern):
    for filesystem in filesystems:
        if fnmatch(filesystem.lower(), pattern.lower()):
            yield filesystem


def get_filesystems_to_snapshot(filesystems, patterns):
    def get_filtered_filesystems(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_filesystems(filesystems, pattern))
        return result

    includes = get_filtered_filesystems(lambda p: p and p[0] != '!')
    excludes = get_filtered_filesystems(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_datastores_to_snapshot(config, datastores):
    patterns = read_datastore_patterns_from_config(config)
    return get_datastores_to_snapshot(datastores, patterns)


def iter_filtered_vms(pattern, datastores, vms):
    for vmname, vmdatastorename, vm in vms:
        for datastorename in vmdatastorename.split(","):
            if datastorename in datastores:
                if fnmatch(vmname.lower(), pattern.lower()):
                    yield vmname, vmdatastorename, vm


def get_vms_to_snapshot(patterns, datastores, vms):
    def get_filtered_vms(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            result.update(iter_filtered_vms(pattern, datastores, vms))
        return result

    includes = get_filtered_vms(lambda p: p and p[0] != '!')
    excludes = get_filtered_vms(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_vms_to_snapshot(config, datastores, vms):
    patterns = read_vm_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def list_of_vms_without_quiesce(config, datastores, vms):
    patterns = read_noquiesce_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def list_of_vms_without_memory(config, datastores, vms):
    patterns = read_nomemory_patterns_from_config(config)
    return get_vms_to_snapshot(patterns, datastores, vms)


def iter_vms(config, datastores, all_vms):
    vms = list_of_vms_to_snapshot(config, datastores, all_vms)
    noquiesce = list_of_vms_without_quiesce(config, datastores, vms)
    nomemory = list_of_vms_without_memory(config, datastores, vms)
    for vm in vms:
        quiesce = vm not in noquiesce
        memory = vm not in nomemory
        yield vm,  quiesce, memory


def is_dryrun(config):
    return 'global' in config and 'dryrun' in config['global'] and config['global']['dryrun']

def is_suppress(config):
    return 'logging' in config and 'suppress' in config['logging'] and config['logging']['suppress']

def is_use_volid(config):
    return 'global' in config and 'use_volid' in config['global'] and config['global']['use_volid']

def is_sa_internal(config):
    return 'global' in config and 'sa_internal' in config['global'] and config['global']['sa_internal']

def is_single_instance(config):
    return 'global' in config and 'single_instance' in config['global'] and config['global']['single_instance']

def is_vmware(config):
    return 'vcenter' in config
