from distutils.version import LooseVersion
from logsetup import UserLogger
from infinisdk import InfiniBox
from infinisdk.core.exceptions import APICommandFailed
from contextlib import contextmanager
import errors
import json
from time import time, gmtime
import pickle
from config import dummy_context_mgr

MAX_SYSTEM_SNAPSHOTS = 100000
MAX_VOL_SNAPSHOTS = 210
MAX_FS_SNAPSHOTS = 4000

MIN_SUPPORTED_VER = LooseVersion("1.5.0")
MAX_SUPPORTED_VER = LooseVersion("3.1.0")
CG_SUPPORTED_VER = LooseVersion("2.2.0")
FS_SUPPORTED_VER = LooseVersion("2.1.0")

log = UserLogger(__name__)


@contextmanager
def infinibox_api_error_wrapper(hostname):
    try:
        yield
    except APICommandFailed, e:
        if e.error_code in ('AUTHENTICATION_REQUIRED', 'UNAUTHORIZED'):
            log.uabort("Could not connect to {} API: username/password error.".format(hostname),
                       exit_code=errors.EXIT_CODE_IBOX_LOGIN_FAILED)
        else:
            log.uabort_with_exc_info(
                "Caught APICommandFailed", exit_code=errors.EXIT_CODE_IBOX_API_FAILED)
    except SystemExit:
        raise
    except Exception, e:
        message = getattr(e, 'message', 'Failed to connect, aborting.')
        log.uabort_with_exc_info(
            message, exit_code=errors.EXIT_CODE_IBOX_NETWORK_ERROR)


def load_cookie(infinibox):
    try:
        with open('/tmp/snaprotator_cookie', 'r') as f:
            return pickle.load(f)
    except:
        log.debug("Cookie not found.")
    return None


def save_cookie(infinibox):
    cookie = None
    try:
        with open('/tmp/snaprotator_cookie', 'w') as f:
            pickle.dump(infinibox.api._session.cookies.copy(), f)
    except:
        log.debug("Failed to save cookie")


def my_login(infinibox):
    cookie = load_cookie(infinibox)
    if not cookie:
        log.debug("Cookie not found. Logging in.")
        infinibox.login()
        cookie = save_cookie(infinibox)
    else:
        log.debug("Loaded saved cookie")
        infinibox.api._session.cookies = cookie
        try:
            infinibox.get_version()
            # It didn't use the cookie, but looged in
            if infinibox.is_logged_in():
                log.debug("Logged in, cookie expired. Saving.")
                save_cookie(infinibox)
        except:
            log.debug("Cookie not accepted")
            infinibox.api._session.cookies.clear()
            infinibox.login()
            save_cookie(infinibox)


def connect_to_infinibox(config, ssl=True):
    hostname = config['infinibox']['host']
    username = config['infinibox']['username']
    password = config['infinibox']['password']

    log.uinfo("Connecting to the Infinibox '{}'".format(hostname))
    with infinibox_api_error_wrapper(hostname):
        infinibox = InfiniBox(hostname, use_ssl=ssl, auth=(username, password))
        my_login(infinibox)
    return infinibox


def report_metadata_stats(infinibox, prefix, sa_internal=False):
    try:
        runs, fs_success, fs_failure, vol_success, vol_failure, warnings = calculate_metadata_stats(
            infinibox, prefix)
        log.info(
            'Total runs for policy "{}" in the last 24 hours: {}, runs with warnings: {}', prefix, runs, warnings)
        log.info('FS success: {} FS failure: {} VOL/CG Success: {} VOL/CG failure: {}',
                 fs_success, fs_failure, vol_success, vol_failure)
        event_24_h(infinibox, runs, prefix, fs_success,
                   fs_failure, vol_success, vol_failure, warnings, sa_internal)
    except:
        log.uwarn("Failed to report metadata stats.")


def calculate_metadata_stats(infinibox, prefix):
    pool = find_smallest_pool(infinibox)
    metadata = get_current_metadata(pool, prefix)
    keys = [x for x in metadata]
    fs_success = 0
    fs_failure = 0
    vol_success = 0
    vol_failure = 0
    warnings = 0
    for entry in keys:
        fs_success += metadata[entry]['data']['fs_success']
        fs_failure += metadata[entry]['data']['fs_failure']
        vol_success += metadata[entry]['data']['vol_success']
        vol_failure += metadata[entry]['data']['vol_failure']
        warnings += metadata[entry]['data']['warnings']

    return len(keys), fs_success, fs_failure, vol_success, vol_failure, warnings


def do_metadata(infinibox, prefix, fs_success, fs_failure, vol_success, vol_failure, warnings, suppress=False, sa_internal=False):
    try:
        pool = find_smallest_pool(infinibox)
        current = fill_current_metadata(
            fs_success, fs_failure, vol_success, vol_failure, warnings)
        set_current_metadata(infinibox, pool, prefix, current, suppress)
        history = get_current_metadata(pool, prefix)
        newest, prev = cleanup_metadata(infinibox, pool, history, suppress)
        log.debug("Newest timestamp: {} Previous timestamp: {}", newest, prev)
        p_year, p_mon, p_mday, p_hour, p_min, p_sec, p_wday, p_yday, p_isdst = gmtime(
            prev)
        n_year, n_mon, n_mday, n_hour, n_min, n_sec, n_wday, n_yday, n_isdst = gmtime(
            newest)
        # Send report if day is different from the previous one
        if n_yday != p_yday or p_year != n_year:
            report_metadata_stats(infinibox, prefix, sa_internal)

    except:
        log.uwarn("Updating metadata failed.")


def fill_current_metadata(fs_success, fs_failure, vol_success, vol_failure, warnings):
    data = dict()
    data['fs_success'] = fs_success
    data['fs_failure'] = fs_failure
    data['vol_success'] = vol_success
    data['vol_failure'] = vol_failure
    data['warnings'] = warnings
    total = dict()
    total['timestamp'] = int(time())
    total['data'] = data
    return total


def find_smallest_pool(infinibox):
    pools = infinibox.pools.to_list()
    return sorted(pools, key=lambda x: x.get_id())[0]


def get_current_metadata(pool, policy_name):
    try:
        all_metadata = pool.get_all_metadata()
        history = {x: json.loads(
            all_metadata[x]) for x in all_metadata if 'snaprotator_' + policy_name + '.' in x}
        return history
    except:
        log.uwarn("No metadata found. Returning default.")
        return dict()


def set_current_metadata(infinibox, pool, policy_name, metadata, suppress=False):
    try:
        meta = json.dumps(metadata)
        timestamp = metadata['timestamp']
        with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
            pool.set_metadata(
                'snaprotator_' + policy_name + '.' + str(timestamp), meta)
    except:
        log.uwarn("Failed to save metadata.")


def cleanup_metadata(infinibox, pool, metadata, suppress=False):
    try:
        keys = sorted([x for x in metadata])
        last_timestamp = metadata[keys[-1]]['timestamp']
        prev_timestamp = last_timestamp
        if len(keys) > 1:
            prev_timestamp = metadata[keys[-2]]['timestamp']
        log.debug("Latest metadata: {}".format(last_timestamp))
        for entry in keys:
            this_timestamp = metadata[entry]['timestamp']
            log.debug("Processing metadata: {}".format(this_timestamp))
            if last_timestamp - this_timestamp > 86400:
                with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
                    pool.unset_metadata(entry)
                log.debug('Successfully deleted metadata {}'.format(entry))
        return last_timestamp, prev_timestamp
    except:
        log.uwarn("Failed to cleanup metadata.")
        return 0, 0


def infinibox_validate(infinibox):
    with infinibox_api_error_wrapper("{}:{}".format(*infinibox.get_api_addresses()[0])):
        infinibox_version = LooseVersion(infinibox.get_version())
        snapshots_count = sum(p.get_field('snapshots_count')
                              for p in infinibox.pools.get_all())

    # Validate constraints
    if not MIN_SUPPORTED_VER < infinibox_version < MAX_SUPPORTED_VER:
        log.uabort("Infinibox version {} is not supported by this version of snaprotator".format(infinibox_version),
                   exit_code=errors.EXIT_CODE_IBOX_VERSION_UNSUPPORTED)
    if snapshots_count >= MAX_SYSTEM_SNAPSHOTS:
        log.uabort("Maximum number of snapshots per system had been reached ({}), rotation will not be performed",
                   MAX_SYSTEM_SNAPSHOTS, exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_SYSTEM)


def volume_validate(infinibox, volume_name):
    log.uinfo("Checking for volume/fs/cg '{}'".format(volume_name))
    infinibox_version = LooseVersion(infinibox.get_version())
    volumes = infinibox.objects.volumes.find(name=volume_name)
    if not volumes:
        # Check if filesystem exists with the same name
        # supported at least since 1.7
        if infinibox_version > FS_SUPPORTED_VER:
            volumes = infinibox.objects.filesystems.find(name=volume_name)
            if not volumes:
                # Check if consistency group exists with the same name
                # supported since v. 2.2
                if infinibox_version > CG_SUPPORTED_VER:
                    volumes = infinibox.objects.cons_groups.find(
                        name=volume_name)
        if not volumes:
            log.uabort("Failed to find volume '{}', aborting.".format(volume_name),
                       exit_code=errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND)

    volume = volumes[0]
    if volume.get_type_name() == 'cons_group':
        snapshots = volume.get_snapgroups()
    else:
        snapshots = volume.get_snapshots()

    if len(snapshots) >= MAX_VOL_SNAPSHOTS:
        log.uabort("Too many snapshots ({}) for volume '{}', aborting.".format(len(snapshots), volume_name),
                   exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_VOLUME)

    log.debug("Volume validation - OK")
    return volume, snapshots


def get_all_volumes(infinibox):
    log.uinfo("Getting all volumes from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    names_of_volumes_cgs = []
    volumes = infinibox.objects.volumes.find(type='MASTER')
    for volume in volumes:
        if infinibox_version > CG_SUPPORTED_VER:
            if volume.get_cons_group() == None:
                if not volume.is_rmr_target():
                    names_of_volumes_cgs.append(volume)
                else:
                    log.debug(
                        "Volume {} is a replication target, skipping".format(volume.get_name()))
        else:
            if not volume.is_rmr_target():
                names_of_volumes_cgs.append(volume)
            else:
                log.debug(
                    "Volume {} is a replication target, skipping".format(volume.get_name()))

    if infinibox_version > CG_SUPPORTED_VER:
        cgs = infinibox.objects.cons_groups.find(type='MASTER')
        for cg in cgs:
            replist = cg.get_replicas().to_list()
            target = False
            if len(replist) > 0:
                for replica in replist:
                    if replica.is_target():
                        log.debug(
                            "CG {} is a replication target, skipping".format(cg.get_name()))
                        target = True
            if not target:
                names_of_volumes_cgs.append(cg)

    if not names_of_volumes_cgs:
        log.uinfo("Failed to find volumes/cgs")
    else:
        log.debug("Volume find - OK")
    return names_of_volumes_cgs


def check_count_fs_snaps(infinibox):
    log.uinfo("Getting all fs snaps from the system")
    volumes = infinibox.objects.filesystems.find(type='SNAP')
    return len(volumes) >= MAX_FS_SNAPSHOTS


def get_all_filesystems(infinibox):
    log.uinfo("Getting all filesystems from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    names_of_filesystems = []
    if infinibox_version > FS_SUPPORTED_VER:
        filesystems = infinibox.objects.filesystems.find(type='MASTER')
        for filesystem in filesystems:
            names_of_filesystems.append(filesystem)

    if not names_of_filesystems:
        log.uwarn("Failed to find filesystems")
    else:
        log.debug("Filesystem find - OK")
    return names_of_filesystems


def send_custom_event(infinibox, level, visibility, description_template, data=None):
    if infinibox is not None:
        infinibox.events.create_custom_event(level=level, description=description_template,
                                             visibility=visibility, data=data)
    else:
        log.debug("CUSTOM_EVENT LEVEL={level} VISIBILITY={visibility} description_template={description_template!r} data={data!r}".format(
            level=level, visibility=visibility, description_template=description_template, data=data))


def event_locked(infinibox, sa_internal=False):
    template = "Another snaprotator process is running"
    from __version__ import __version__ as snaprotator_version
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"
    data = [
        dict(name="snaprotator_version", type="string", value=snaprotator_version),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
    ]
    send_custom_event(infinibox, level="WARNING",
                      visibility="CUSTOMER", description_template=template, data=data)


def event_24_h(infinibox, runs, policy, fs_success, fs_failure, vol_success, vol_failure, warnings, sa_internal=False):
    level = "INFO"
    if vol_failure > 0 or fs_failure > 0:
        level = "ERROR"
    elif warnings:
        level = "WARNING"
    from __version__ import __version__ as snaprotator_version
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"

    template = 'Snaprotator 24h status for policy "{policy}": {level}. Total {runs} runs. FS success: {fs_success} FS failure: {fs_failure} VOL/CG Success: {vol_success} VOL/CG failure: {vol_failure}. Runs with warnings: {warnings}'
    data = [
        dict(name="policy", type="string", value=policy),
        dict(name="runs", type="string", value=runs),
        dict(name="level", type="string", value=level),
        dict(name="warnings", type="string", value=warnings),
        dict(name="fs_success", type="string", value=fs_success),
        dict(name="fs_failure", type="string", value=fs_failure),
        dict(name="vol_success", type="string", value=vol_success),
        dict(name="vol_failure", type="string", value=vol_failure),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
        dict(name="snaprotator_version", type="string", value=snaprotator_version),
    ]
    send_custom_event(infinibox, level=level,
                      visibility="CUSTOMER", description_template=template, data=data)


def event_unknown_error(infinibox, sa_internal=False):
    template = "Snaprotator finished with status: ERROR"
    from __version__ import __version__ as snaprotator_version
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"
    data = [
        dict(name="snaprotator_version", type="string", value=snaprotator_version),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
    ]
    send_custom_event(infinibox, level="ERROR",
                      visibility="CUSTOMER", description_template=template, data=data)
