import os
import sys
import logbook
from logbook.more import ColorizedStderrHandler
from errors import EXIT_CODE_GENERAL_ERROR

_has_warnings = False


def has_warnings():
    return _has_warnings


class UserLogger(logbook.Logger):

    """docstring for UserLogger"""

    def uinfo(self, *args, **kwargs):
        self.info(*args, user=True, **kwargs)

    def uwarn(self, *args, **kwargs):
        global _has_warnings
        _has_warnings = True
        self.warn(*args, user=True, **kwargs)

    def uabort(self, *args, **kwargs):
        self.error(*args, user=True, **kwargs)
        exit_code = kwargs.get('exit_code', EXIT_CODE_GENERAL_ERROR)
        raise SystemExit(exit_code)

    def uabort_with_exc_info(self, *args, **kwargs):
        self.error(*args, user=True, exc_info=True, **kwargs)
        exit_code = kwargs.get('exit_code', EXIT_CODE_GENERAL_ERROR)
        raise SystemExit(exit_code)

log = UserLogger(__name__)


def setup_file_handler(config):
    try:
        file_path = config['logging']['logpath']
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))
        if 'norotation' in config['logging']:
            file_handler = logbook.MonitoringFileHandler(
                file_path, level=logbook.DEBUG, bubble=True)
        else:
            date_format = '%Y-%m-%d'
            file_handler = logbook.TimedRotatingFileHandler(
                file_path, mode='a', date_format=date_format, backup_count=14, level=logbook.DEBUG, bubble=True)

    except Exception:
        msg = "caught exception when creating logfile {}".format(file_path)
        err = sys.exc_info()[1]
        sys.stderr.write('{msg} ({err})\n'.format(msg=msg, err=err))
        log.uabort_with_exc_info(msg)
    return file_handler


def setup_email(config):
    class MyMailHandler(logbook.MailHandler):

        def emit_batch(self, records, _):
            try:
                if not records:
                    return
                max_level = max(r.level for r in records)
                if max_level >= logbook.ERROR:
                    self.subject = "ERROR: snaprotator: Failed"
                elif max_level >= logbook.WARNING:
                    self.subject = "WARNING: snaprotator: Some operations failed"
                else:
                    self.subject = "SUCCESS: snaprotator: Finished successfully"
                return super(MyMailHandler, self).emit_batch(records, 'escalation')
            except:
                err = sys.exc_info()[1]
                sys.stderr.write(
                    'Error sending mail. Please check your configuration file. ({})\n'.format(err))

    mail_format_string = u'''\
    Subject: {handler.subject}

    Message type:       {record.level_name}
    Location:           {record.filename}:{record.lineno}
    Module:             {record.module}
    Function:           {record.func_name}
    Time:               {record.time:%Y-%m-%d %H:%M:%S}

    Message:

    {record.message}
    '''

    mail_related_format_string = u'{record.time:%Y-%m-%d %H:%M:%S}:{record.level_name}:{record.message}'

    sender = config['logging']['emailsender']
    rcpt = [r for r in config['logging']['emailrcpt'].split(',')]
    emailserver = config['logging']['emailserver']

    if 'emailport' in config['logging']:
        emailport = int(config['logging']['emailport'])
    else:
        emailport = 25

    kargs = {}
    kargs['from_addr'] = sender
    kargs['recipients'] = rcpt
    kargs['server_addr'] = (emailserver, emailport)
    kargs['level'] = 'INFO'
    kargs['format_string'] = mail_format_string
    kargs['related_format_string'] = mail_related_format_string
    kargs['bubble'] = True

    if 'emailuser' in config['logging']:
        username = config['logging']['emailuser']
        password = config['logging']['emailpassword']
        kargs['credentials'] = (username, password)
        kargs['secure'] = ()

    email_handler = MyMailHandler(**kargs)
    return email_handler


def setup_logging(config):
    logbook.set_datetime_format('local')
    loggers = []

    class NoExceptionStringFormatter(logbook.StringFormatter):

        def format_exception(self, _):
            return None

    console_level = logbook.INFO
    console_filter = lambda r, _: r.kwargs.get('user', False)
    console_log_format = (
        '{record.time:%H:%M}\t'
        '{record.level_name}\t'
        '{record.message}'
    )
    if 'logging' in config:
        # Set up logging levels
        if 'debug' in config['logging'] and config['logging']['debug']:
            console_level = logbook.DEBUG
            console_filter = lambda r, _: True
            console_log_format = logbook.StringFormatterHandlerMixin.default_format_string
        elif 'quiet' in config['logging'] and config['logging']['quiet']:
            console_level = logbook.ERROR
            console_log_format = '{record.message}'

        # Add logfile to handlers if requested
        if 'logpath' in config['logging']:
            file_handler = setup_file_handler(config)
            loggers.append(file_handler)

        if 'syslog' in config['logging'] and config['logging']['syslog']:
            syslog_handler = logbook.SyslogHandler(
                application_name=log.name,
                level=logbook.DEBUG,
                bubble=True
            )
            loggers.append(syslog_handler)

    stderr_handler = ColorizedStderrHandler(
        level=console_level,
        filter=console_filter,
        bubble=True
    )
    stderr_handler.formatter = NoExceptionStringFormatter(console_log_format)
    loggers.append(stderr_handler)

    return logbook.NestedSetup(loggers)


def setup_mail_logging(config, dryrun):
    if not dryrun and ('logging' in config and 'email' in config['logging'] and config['logging']['email']):
        email_handler = setup_email(config)
        return logbook.GroupHandler(email_handler)
    else:
        return logbook.NullHandler()
