import json
import requests
import sys
import vmware
from logsetup import UserLogger

requests.packages.urllib3.disable_warnings()

import ssl
try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context

log = UserLogger(__name__)


def http(url, data):
    log.debug('http: url  {}', url)
    log.debug('http: data {}', data)
    try:
        r = requests.post(url, data=data, allow_redirects=True, timeout=60)
        return True, r.content
    except:
        e = sys.exc_info()[0]
        log.uwarn('Failed HTTP request. url: {}, data: {}, err: {}', url, data, e)
        return False, e


def delete_snapshot(powertools_info, vm_refs, snap_uuid, vm_uuid):
    host_power_tools_vm_ip, session_key = powertools_info
    vm_ref_list = vm_refs.split(',')
    return_error = None
    for vm_ref in vm_ref_list:
        log.debug('delete_snapshot: Trying VM {}', vm_ref)
        url = 'http://%s/plugin/views/virtualmachine/operations/remove_snapshot?moref=%s&uuid=%s&locale=en_US&vimSessionKey=%s' %\
            (host_power_tools_vm_ip, vm_ref, vm_uuid, session_key)
        data = {"snapshot_id": snap_uuid}
        success, resp = http(url, data)

        if not success:  # http request failed, we don't need to retry
            return False, resp
        json_response = {}
        try:
            json_response = json.loads(resp)
        except:
            return_error = str(sys.exc_info()[0])
            log.debug('Failed to delete snapshot from Host Power Tools for VMWare {}', return_error)
            log.debug(resp)
        if 'success' in json_response and json_response['success']:
            return True, json_response
    log.uwarn(
        'Tried all VMs, still can\'t delete snapshot from Host Power Tools for VMWare. Please delete the snapshot manually.')
    return False, return_error


def delete_snapshot_with_metadata(powertools_info, metadata):
    original_datastore = None
    snapshot_name = None
    if 'original_datastore' in metadata and 'name' in metadata:
        original_datastore = metadata['original_datastore']
        snapshot_name = metadata['name']

    if 'vm_refs' in metadata and 'first_vm_uuid' in metadata and 'first_vm_snapshot_id' in metadata:
        vm_refs = metadata['vm_refs']
        snap_uuid = metadata['first_vm_snapshot_id']
        vm_uuid = metadata['first_vm_uuid']
        success, myerror = delete_snapshot(powertools_info,
                                           vm_refs, snap_uuid, vm_uuid)
        if success:
            log.uinfo('Successfuly removed snapshot ({}) from VMware Power Tools (ds: {})',
                      snapshot_name, original_datastore)
            return True, "Success"
        else:
            log.uwarn('Failed to remove snapshot ({}) from VMware Power Tools (ds: {})({})',
                      snapshot_name, original_datastore, myerror)
            return False, myerror

    return True, "Can't find supported metadata.Either no vms found on datastore or check VMWare Host Power Tools version."


def get_all_vms_string(vms):
    vmstring = ''
    for vm in vms:
        if vm.summary.config.uuid is None:
            continue
        vmstring = vmstring + vmware.get_vm_string_name(vm) + ";"
    return vmstring


def get_all_vms_names(vms):
    return [vm.name for vm in vms]


def register_snapshot(powertools_info, vms_infinisnapid_datastore):
    host_power_tools_vm_ip, session_key = powertools_info
    vms, infinisnapid, datastore = vms_infinisnapid_datastore
    datastorename = vmware.get_datastore_string_name(datastore)
    vmstring = get_all_vms_string(vms)
    if vmstring == '':
        return True, "No vms found on datastore"
    vmnames = get_all_vms_names(vms)

    url = 'http://%s/plugin/views/datastore/register_snapshot?moref=%s&locale=en_US&vimSessionKey=%s' %\
        (host_power_tools_vm_ip, datastorename, session_key)
    data = {"snapshot_id": infinisnapid, "vmrefs": vmstring}
    success, resp = http(url, data)
    if not success:
        return False, "HTTP request failed"

    try:
        json_response = json.loads(resp)
    except:
        e = sys.exc_info()[0]
        log.uwarn('Failed to register snapshot with Host Power Tools for VMware {}', str(e))
        return False, resp
    if 'success' not in json_response or not json_response['success']:
        if 'error_message' in json_response:
            err = json_response['error_message']
        else:
            err = json_response
        log.uwarn('Failed to register snapshot with Host Power Tools for VMware {} ({})', datastore.name, str(err))
        return False, err
    else:
        log.uinfo('Successfully registered snapshot with Host Power Tools for VMware for {}', datastore.name)
        log.uinfo('Registered VMS: {}', vmnames)
        return True, 'Success'


def get_powertools_info(client):
    host_power_tools_vm_ip = vmware.find_vmware_power_tools_ip(client)
    session_key = client.session_manager.currentSession.key
    if host_power_tools_vm_ip and session_key:
        return (host_power_tools_vm_ip, session_key)
    else:
        return None


def register_snap_unregister_snaps(powertools_info, vms_infinisnapid_datastore, deleted_snaps_metadata_list):
    success, err = register_snapshot(powertools_info, vms_infinisnapid_datastore)
    if success:
        log.debug(err)
    else:
        log.uwarn('Failed to register. ({})', str(err))

    for metadata in deleted_snaps_metadata_list:
        success, err = delete_snapshot_with_metadata(powertools_info, metadata)
        if success:
            log.debug(err)
        else:
            log.uwarn('Failed to unregister. ({})', str(err))
