import re
from logsetup import UserLogger
from config import dummy_context_mgr

log = UserLogger(__name__)


def filter_irrelevant_snapshots(snapshots, volume, basename):
    def pred(snapshot):
        name = snapshot.get_name()
        prefix = r'^.*\.{}\.'.format(re.escape(basename))
        date = r'\d{4}\.\d{2}\.\d{2}_\d{6}$'  # yyyy.mm.dd_hhmmss
        return bool(re.match(prefix + date, name))
    return [s for s in snapshots if pred(s)]


def make_snapshot_name(volume, basename, time, uuid):
    from datetime import datetime
    dt = datetime.fromtimestamp(time)
    if uuid == True:
        name = volume.get_id()
    else:
        name = volume.get_name()
    return "{}.{}.{}".format(name, basename, dt.strftime("%Y.%m.%d_%H%M%S"))


def sort_and_filter_snapshots(snapshots, volume, basename):
    """:return: sorted list of snapshots where snapshots
    not conforming to the name template are removed"""
    return sorted(filter_irrelevant_snapshots(snapshots, volume, basename),
                  key=lambda o: o.get_name())


def trim_number_of_snapshots(infinibox, snapshots, volume, basename, retain, dryrun, suppress=False):
    """We try to delete as many snapshots as needed to keep the number of snapshots less or equal to self.retain."""
    snapshots = sort_and_filter_snapshots(snapshots, volume, basename)
    num_to_delete = max(0, len(snapshots) - retain)
    snapshots_to_delete = snapshots[:num_to_delete]
    metadata = list()
    for snap in snapshots_to_delete:
        metadata.append(snap.get_all_metadata())
        name = None
        try:
            name = snap.get_name()
            if not dryrun:
                if snap.get_type_name() == 'cons_group':
                    with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS" : "true"}) if suppress else dummy_context_mgr():
                        snap.delete(delete_members=True)
                    log.uinfo("Deleted snapshot {}".format(name))
                else:
                    if snap.has_children() == False:
                        with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS" : "true"}) if suppress else dummy_context_mgr():
                            snap.delete()
                        log.uinfo("Deleted snapshot {}".format(name))
                    else:
                        log.uwarn("Clone exists for snapshot {}. Not deleting.".format(name))
        except:
            log.uwarn(
                "Failed to delete snapshot {}, moving on".format(name), exc_info=True)
    return metadata


def create_latest_snapshot(infinibox, volume, basename, dryrun, uuid, suppress=False):
    """ Create latest snapshot <volname>.<basename>.YYYY.MM.DD_HHMMSS"""
    from time import time
    snapshot = None
    try:
        name = make_snapshot_name(volume, basename, time(), uuid)
        voltype = volume.get_type_name()
        log.uinfo(
            "Creating snapshot {} for {} '{}'".format(name, voltype, volume.get_name()))
        if dryrun:
            return None
        with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS" : "true"}) if suppress else dummy_context_mgr():
            snapshot = volume.create_snapshot(name)
        return snapshot
    except:
        log.uwarn("Failed to create latest snapshot", exc_info=True)
        return None
