"""

Usage:
   snaprotator [options]
   snaprotator -h | --help
   snaprotator --version


Options:
   -c configfile --configfile=<configfile>        Config file location
   -H hostname --hostname=<hostname>              DNS name of the infinibox
   -u user --user=<username>                      Administrator's username
   -p password --password=<password>              Administrator's password
   -v volume1[,volume2...] --volumes=<volume1[,volume2...>
                                                  Volume(s) to rotate the snaphsots on
   -b basename --basename=<basename>              Basename (daily,weekly, etc...)
   -r retain --retain=<retain>                    Number of snapshots to retain
   -l logfile --logfile=<logfile>                 Use log file instead of stderr
   -h, --help                                     Show this screen
   -D, --debug                                    Debug
   -q, --quiet                                    Suppress all informational messages
   -d, --dryrun                                   Show what will be done
   -P, --purge                                    Purge all snapshots with basename <basename>
   -s, --syslog                                   Log to syslog daemon
   -n, --norotation                               Do NOT rotate logfile, keep writing indefinitely
   -V, --version                                  Show snaprotator version
"""
import signal
import sys
import logbook
from logsetup import UserLogger, setup_logging, setup_mail_logging, has_warnings

from docopt import docopt
from __version__ import __version__ as snaprotator_version
from rotation import (create_latest_snapshot, trim_number_of_snapshots)

import vmware
import config as conf
from infinibox import do_metadata, connect_to_infinibox, infinibox_validate, volume_validate, get_all_volumes, check_count_fs_snaps, get_all_filesystems, event_unknown_error, event_locked
from infi.pyvmomi_wrapper import Client
from time import sleep, time
import powertools
import errors
from infi.pyutils.contexts import contextmanager

log = UserLogger(__name__)

MAX_ATTEMPTS = 3
RETRY_IN = 5
MAX_RETAIN = 250
DEFAULT_PARALLEL = 5


def timed(func):
    def wrapper(*arg, **kw):
        t1 = time()
        res = func(*arg, **kw)
        t2 = time()
        took = t2 - t1
        tooks = int(took)
        tookms = int(took * 1000 - tooks)
        log.debug("{} took {}.{} seconds".format(func.__name__, tooks, tookms))
        return res
    return wrapper


@contextmanager
def locking_context(name):
    from lockfile import AlreadyLocked
    from lockfile.pidlockfile import PIDLockFile
    from os import path
    lock = PIDLockFile(name, timeout=0)

    while True:
        try:
            if lock.i_am_locking():  # our PID is already acquired the lock
                log.uinfo(
                    "snaprotator process already holds the pid lock file")
                break
            lock.acquire(0)
            log.debug("Lock acquired successfully")
            break  # successfully acquired the lock
        except AlreadyLocked:
            locking_pid = lock.read_pid()  # either None or int
            if locking_pid is not None and path.exists("/proc/{}".format(locking_pid)):
                log.uabort_with_exc_info("another snaprotator process (pid {}) is still running and did not release the lock file"
                                         .format(locking_pid), exit_code=errors.EXIT_CODE_ALREADY_RUNNING)
            # pid file is left over by a dead process, break the lock and
            # re-acquire
            log.uinfo("detected lock left over of a dead snaprotator process ({}), breaking lock".format(
                locking_pid))
            lock.break_lock()

    try:
        yield
    finally:
        lock.release()


def signal_handler(signal, frame):
    msg = (
        'Execution has been interrupted by Ctrl+C!\n' +
        'Please re-run the tool, to make sure the rotation is in order'
    )
    log.uwarn(msg)
    sys.exit(errors.EXIT_CODE_INTERRUPT)


@timed
def connect_to_vcenter(config):
    from requests.exceptions import (
        ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError)
    from pyVmomi import vim

    hostname = config['vcenter']['host']
    username = config['vcenter']['username']
    password = config['vcenter']['password']

    log.uinfo("Connecting to VCenter '{}'".format(hostname))
    try:
        client = Client(hostname, username=username, password=password)
    except (ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError) as e:
        log.uabort_with_exc_info("VCenter connection error, aborting. {}", str(e),
                                 exit_code=errors.EXIT_CODE_VMWARE_NETWORK_ERROR)
    except vim.fault.InvalidLogin as e:
        log.uabort_with_exc_info("VCenter login error, aborting. {}", e.msg,
                                 exit_code=errors.EXIT_CODE_VMWARE_LOGIN_FAILED)
    except SystemExit:
        raise
    except:
        e = sys.exc_info()[0]
        log.uabort_with_exc_info("Failed to connect to VCenter, aborting. Unknown error: {}", str(e),
                                 exit_code=errors.EXIT_CODE_VMWARE_GENERAL_ERROR)

    return client


def pipeline_vm_tasks(parallel, items, task_factory, task_handler):
    tasks = dict()
    while items or tasks:
        if len(tasks) < parallel and items:
            item = items.pop()
            task = task_factory(item)
            if task is not None:
                tasks[item] = task

        for item, task in tasks.copy().iteritems():
            if task_handler(item, task):
                del tasks[item]
        sleep(0.1)


@timed
def do_infinibox_snapshot(infinibox, basename, retain, volume_name, dryrun, uuid=False, suppress=False):
    volume, snapshots = volume_validate(infinibox, volume_name)
    snapshot_obj = create_latest_snapshot(
        infinibox, volume, basename, dryrun, uuid, suppress)
    deleted_snaps_metadata = trim_number_of_snapshots(
        infinibox, snapshots, volume, basename, retain - 1, dryrun, suppress)
    return snapshot_obj, deleted_snaps_metadata


@timed
def iter_filtered_datastores_volumes(client, config, infinibox, datastores):
    log.debug("iter_filtered_datastores_volumes")
    filtered_datastores = conf.list_of_datastores_to_snapshot(
        config, datastores)
    log.debug("conf.list_of_datastores_to_snapshot")
    infi_datastores = vmware.iter_filtered_datastores_block_devices(
        client, filtered_datastores)
    log.debug("vmware.iter_filtered_datastores_block_devices")
    for datastore, infinidat_devices in infi_datastores:
        for device in infinidat_devices:
            volume = infinibox.objects.volumes.find(serial=device)
            if volume:
                yield datastore, datastore.name, datastore.vm, volume


def get_datastore_infinibox_volumes(client, config, infinibox, datastores):
    return list(iter_filtered_datastores_volumes(client, config, infinibox, datastores))


def get_datastore_snapshot(infinibox_snapshot, datastores_volumes):
    for datastore, datastore_name, datastore_vms, volume_list in datastores_volumes:
        if infinibox_snapshot.get_parent() in volume_list:
            return datastore, datastore_name, datastore_vms, infinibox_snapshot


def get_vms_infinisnapid_datastore(datastores_snapshot):
    datastore, datastore_names, datastore_vms, infinisnap = datastores_snapshot
    vms = vmware.get_all_vms_on_datastore_with_single_datastore(datastore)
    return vms, infinisnap.get_id(), datastore


def get_vms_in_datastore_within_infinisnapid(datastore_volumes, infinibox_snapshot, datastores):
    datastore_snapshot = get_datastore_snapshot(
        infinibox_snapshot, datastore_volumes)
    log.debug("Finished get_datastore_snapshot")
    return get_vms_infinisnapid_datastore(datastore_snapshot)


@timed
def parallel_create_vmware_snapshots(client, parallel, vms, basename):
    vmware_snapshots = []

    def create_snap_task_factory(vm_tuple):
        vm, quiesce, memory = vm_tuple
        return vmware.initiate_snapshot_for_vm(client, basename, vm, quiesce=quiesce, memory=memory)

    def create_snap_task_handler(vm_tuple, task):
        try:
            vm = vm_tuple[0]
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                snapshot = vmware.wait_for_result(task)
                vmware_snapshots.append((vm, snapshot))
                log.uinfo(
                    'Successfully created snapshot for VM {} ({})', vm.name, snapshot)
            elif state == 'error':
                log.uwarn(
                    'Snapshot creation failed. VM Name: {}, reason: {}', vm.name, message)
        except:
            log.uwarn('Unexpected error in create snap', exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vms, create_snap_task_factory, create_snap_task_handler)
    return vmware_snapshots


@timed
def parallel_remove_vmware_snapshots(parallel, vmware_snapshots):
    def remove_snap_task_factory(vm_snapshot_pair):
        return vmware.initiate_remove_snapshot(vm_snapshot_pair)

    def remove_snap_task_handler(vm_snapshot_pair, task):
        try:
            vm, snapshot = vm_snapshot_pair
            state, message = vmware.get_task_status(task)
            if state == 'success':  # success
                log.uinfo(
                    'Successfully deleted snapshot {} for VM {}', snapshot, vm.name)
            elif state == 'error':
                log.uwarn(
                    'Snapshot deletion failed. VM Name: {}, reason: {}', vm.name, message)
        except:
            log.uwarn('Unexpected error in remove snap', exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vmware_snapshots, remove_snap_task_factory, remove_snap_task_handler)


@timed
def cleanup_vmware_snapshots(vms, client, basename):
    for vm, quiesce, memory in vms:
        try:
            log.debug('Cleanup VMWare snapshots for host {}', vm.name)
            vmware.cleanup_infinidat_snapshots(client, basename, vm)
        except:
            log.uwarn('Failed to cleanup snapshot {}', str(sys.exc_info()[0]))


@timed
def register_snap_unregister_snaps(
        powertools_info, vms_infinisnapid_datastore, metadata_list):
    return powertools.register_snap_unregister_snaps(
        powertools_info, vms_infinisnapid_datastore, metadata_list)


@timed
def do_infinibox_vmware_snapshots(client, config, parallel, retain, infinibox, basename, dryrun, suppress=False):
    infini_datastores = vmware.get_all_infinidat_datastores(client)
    datastores_to_snapshot = get_datastore_infinibox_volumes(
        client, config, infinibox, infini_datastores)
    if not datastores_to_snapshot:
        log.uabort("Can't find datastores to snapshot. Exiting.",
                   exit_code=errors.EXIT_CODE_VMWARE_NO_DATASTORES_FOUND)
    all_vms = list(vmware.iter_all_vms(client))
    use_volid = conf.is_use_volid(config)
    powertools_info = None
    todo_volumes = 0
    done_volumes = 0
    if not dryrun:
        powertools_info = powertools.get_powertools_info(client)

    for datastore_volume_tuple in datastores_to_snapshot:
        # zip(*datastore_volume_tuple)
        datastore, datastore_name, datastore_vms, infinibox_volumes = datastore_volume_tuple
        # datastore, datastore.name, datastore.vm, volume
        # datastore_names = [datastore.name for datastore in datastores]
        log.uinfo('Datastore to snap: {}', datastore_name)

        vms_tuple = list(conf.iter_vms(config, [datastore_name], all_vms))
        vms = []

        if not vms_tuple:
            log.uwarn(
                'No VMs found. Continuing infinibox snapshots without OS consistency.')
            vms_names = []
        else:
            vms_names = [vmname for vmname, vmdatastore,
                         vm in [vm for vm, quiesce, memory in vms_tuple]]
            for vmobj, quiesce, memory in vms_tuple:
                vmname, vmdatastore, vm = vmobj
                res = vm, quiesce, memory
                vms.append(res)
        log.uinfo('VMS to snap: {}', [str(a) for a in vms_names])
        log.debug("infinibox volumes: {}".format(
            [vol for vol in infinibox_volumes]))
        volume_names = [volume.get_name()
                        for vlist in [infinibox_volumes] for volume in vlist]
        todo_volumes += len(volume_names)

        log.uinfo(
            "Infinibox volumes to snap: {}", [str(a) for a in volume_names])
        if not dryrun:
            vmware_snapshots = parallel_create_vmware_snapshots(
                client, parallel, list(vms), basename)

        for infinibox_volume in volume_names:
            try:
                done_volumes += 1
                snapshot_name, metadata_list = do_infinibox_snapshot(
                    infinibox, basename, retain, infinibox_volume, dryrun, use_volid, suppress)
                if not snapshot_name:
                    done_volumes -= 1

            except SystemExit, e:
                done_volumes -= 1
                if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                    log.uwarn(
                        "Infinibox volume '{}' not found, skipping it.", infinibox_volume)
                    continue
                else:
                    raise
            if not dryrun:
                if powertools_info:
                    log.debug(
                        "Starting get_vms_in_datastore_within_infinisnapid")
                    vms_infinisnapid_datastore = get_vms_in_datastore_within_infinisnapid(
                        datastores_to_snapshot, snapshot_name, infini_datastores)
                    log.debug("Found datastores")
                    register_snap_unregister_snaps(
                        powertools_info, vms_infinisnapid_datastore, metadata_list)
                    log.debug("Finished register")
                else:
                    log.uinfo(
                        'No Host Power Tools For VMWare found. Check appliance installed/powered on.')

        if not dryrun:
            parallel_remove_vmware_snapshots(parallel, vmware_snapshots)
            cleanup_vmware_snapshots(vms, client, basename)
    return todo_volumes, done_volumes


@timed
def do_purge(infinibox, basename, volume_name, dryrun, suppress=False):
    volume, snapshots = volume_validate(infinibox, volume_name)
    deleted_snaps_metadata = trim_number_of_snapshots(
        infinibox, snapshots, volume, basename, 0, dryrun, suppress)
    # fixme add vmware deletion
    return deleted_snaps_metadata


@timed
def wrapper_get_all_volumes(infinibox):
    return get_all_volumes(infinibox)


@timed
def wrapper_get_all_filesystems(infinibox):
    return get_all_filesystems(infinibox)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    arguments = docopt(__doc__, version=snaprotator_version)

    config = conf.load_config(arguments) if arguments['--configfile'] else {}
    conf.update_arguments_to_config(arguments, config)
    if not config:
        docopt(__doc__, '-h')

    validate, error = conf.validate_config(config)
    if not validate:
        sys.stderr.write(
            'Configuration file or command line parameters error: {}\n'.format(error.errors))
        sys.exit(errors.EXIT_CODE_CONFIG_ERROR)

    log_handler = setup_logging(config)

    basename = config['global']['snap_basename']
    retain = config['global']['snaps_to_keep']

    if 'purge' in config['infinibox']:
        purge = config['infinibox']['purge']
    else:
        purge = False
    dryrun = conf.is_dryrun(config)
    suppress = conf.is_suppress(config)
    is_vmware = conf.is_vmware(config)
    use_volid = conf.is_use_volid(config)
    single_instance = conf.is_single_instance(config)
    sa_internal = conf.is_sa_internal(config)
    mail_handler = setup_mail_logging(config, dryrun)
    infinibox = None
    done_vol_count = 0
    done_fs_count = 0
    todo_vol_count = 0
    todo_fs_count = 0
    with logbook.NullHandler().applicationbound(), mail_handler.applicationbound():
        with log_handler.applicationbound():
            try:
                if dryrun:
                    log.uinfo('DRYRUN')
                infinibox = connect_to_infinibox(config)
                infinibox_validate(infinibox)
                try:
                    with locking_context("/tmp/snaprotator.pid") if single_instance else conf.dummy_context_mgr():
                        # volumes key must exist because validate_config
                        # ensured that
                        if not is_vmware:
                            if 'volumes' in config['infinibox']:
                                standalone_volumes = config[
                                    'infinibox']['volumes'].split(',')
                            else:
                                all_volume_objects = wrapper_get_all_volumes(
                                    infinibox)
                                all_volumes = [vol.get_name()
                                               for vol in all_volume_objects]
                                standalone_volumes = conf.list_of_volumes_to_snapshot(
                                    config, all_volumes)

                            log.uinfo(
                                'The following volumes will be snapshotted: {}', [str(a) for a in standalone_volumes])
                            all_filesystem_objects = wrapper_get_all_filesystems(
                                infinibox)

                            all_filesystems = [fs.get_name()
                                               for fs in all_filesystem_objects]
                            standalone_filesystems = conf.list_of_filesystems_to_snapshot(
                                config, all_filesystems)
                            log.uinfo(
                                'The following filesystems will be snapshotted: {}', [str(a) for a in standalone_filesystems])
                            if not standalone_filesystems and not standalone_volumes:
                                log.uabort_with_exc_info(
                                    "No filesystems/volumes/cgs found. Aborting.")
                            todo_fs_count = len(standalone_filesystems)
                            todo_vol_count = len(standalone_volumes)
                            log.debug(
                                "Total of {} fs, {} volumes to snap".format(todo_fs_count, todo_vol_count))
                            if check_count_fs_snaps(infinibox):
                                total_snaps = standalone_filesystems * retain

                        if purge:
                            for volume in standalone_volumes:
                                do_purge(infinibox, basename, volume, dryrun)
                            for filesystem in standalone_filesystems:
                                do_purge(
                                    infinibox, basename, filesystem, dryrun)
                        elif is_vmware:
                            client = connect_to_vcenter(config)
                            parallel = config['vcenter']['parallel'] if 'parallel' in config[
                                'vcenter'] else DEFAULT_PARALLEL
                            todo_vol_count, done_vol_count = do_infinibox_vmware_snapshots(
                                client, config, parallel, retain, infinibox, basename, dryrun, suppress)
                        else:  # infinibox only
                            done_vol_count = todo_vol_count
                            for volume in standalone_volumes:
                                try:
                                    snapshot_name, metadata_list = do_infinibox_snapshot(
                                        infinibox, basename, retain, volume, dryrun, use_volid, suppress)
                                    if not snapshot_name:
                                        done_vol_count -= 1
                                except SystemExit, e:
                                    done_vol_count -= 1
                                    if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                                        log.uwarn(
                                            'Infinibox volume {} not found, skipping it.', volume)
                                        continue
                                    else:
                                        raise
                            done_fs_count = todo_fs_count
                            for filesystem in standalone_filesystems:
                                try:
                                    snapshot_name, metadata_list = do_infinibox_snapshot(
                                        infinibox, basename, retain, filesystem, dryrun, use_volid, suppress)
                                    if not snapshot_name:
                                        done_fs_count -= 1
                                except SystemExit, e:
                                    done_fs_count -= 1
                                    if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                                        log.uwarn(
                                            'Infinibox filesystem {} not found, skipping it.', filesystem)
                                        continue
                                    else:
                                        raise
                        if has_warnings():
                            log.uinfo('Done (with warnings). FS: {} out of {}, Vol: {} out of {}'.format(
                                done_fs_count, todo_fs_count, done_vol_count, todo_vol_count))
                            raise SystemExit(errors.EXIT_CODE_WARNINGS_FOUND)
                        else:
                            log.uinfo('Finished without warnings or errors. FS: {} out of {}, Vol: {} out of {}'.format(
                                done_fs_count, todo_fs_count, done_vol_count, todo_vol_count))
                            if not dryrun:
                                do_metadata(infinibox, basename, done_fs_count, todo_fs_count -
                                            done_fs_count, done_vol_count, todo_vol_count - done_vol_count, 0, suppress, sa_internal)
                    log.info("Done.")
                except SystemExit, e:
                    if(e.code == errors.EXIT_CODE_ALREADY_RUNNING):
                        raise SystemExit(errors.EXIT_CODE_ALREADY_RUNNING)
                    elif(e.code == errors.EXIT_CODE_WARNINGS_FOUND):
                        raise SystemExit(errors.EXIT_CODE_WARNINGS_FOUND)
                    else:
                        raise SystemExit(errors.EXIT_CODE_GENERAL_ERROR)
                except:
                    raise SystemExit(errors.EXIT_CODE_GENERAL_ERROR)

            except SystemExit, e:
                if not dryrun:
                    if(e.code == errors.EXIT_CODE_WARNINGS_FOUND):
                        do_metadata(infinibox, basename, done_fs_count, todo_fs_count -
                                    done_fs_count, done_vol_count, todo_vol_count - done_vol_count, 1, suppress, sa_internal)
                    elif (e.code == errors.EXIT_CODE_ALREADY_RUNNING):
                        event_locked(infinibox, sa_internal)
                    else:
                        event_unknown_error(infinibox, sa_internal)
                log.uwarn("Done with warnings/errors.")
                sys.exit(e.code)
            except:
                log.debug("Unknown Error")
                if not dryrun:
                    event_unknown_error(infinibox, sa_internal)
                log.uabort_with_exc_info(
                    "Internal error occurred, please contact Infinidat support.")
