from contextlib import contextmanager
from infinisim import Infinibox as Simulator
from infinisdk import InfiniBox
from capacity import GiB, TiB
from unittest import TestCase
import config as conf
import re
import logbook
import vmware
import os
import rotation
import snaprotator
import errors
import infinibox
from mock import Mock


class TestRotation(TestCase):
    VOL_PREFIX = 'testvol'
    CREATION_DATES = ['2014.10.22_150809', '2014.10.22_160801',
                      '2014.10.22_152931', '2014.10.22_180801',
                      '2014.10.22_130104', '2014.10.22_010801',
                      '2014.10.22_050809', '2014.10.21_160801']
    BASENAMES = ['daily', VOL_PREFIX]
    MAXVOLS = 5
    NAME_TEMPLATE = "{volname}.{basename}.{date}"
    TEST_NAME_TEMPLATES = [NAME_TEMPLATE,
                           "{volname}.{basename}.{date}.{date}",
                           "{volname}.{basename}.{volname}",
                           "{volname}.{date}.{date}", "{volname}.{basename}",
                           "{volname}.{date}", "{volname}.{volname}"]

    def setUp(self):
        self.test_log = logbook.TestHandler()
        self.test_log.push_application()
        # Redefine Globals - prorably should be done with fixtures per test
        infinibox.MAX_SYSTEM_SNAPSHOTS = 500
        infinibox.MAX_VOL_SNAPSHOTS = 100
        infinibox.MAX_VOL_NAME_LEN = 46
        infinibox.MAX_RETAIN = 100

    @contextmanager
    def assert_exits_with_error(self, re_str, exit_code):
        try:
            yield
        except SystemExit, e:
            self.assertTrue(self.test_log.has_error(re.compile(re_str)), self.test_log.formatted_records)
            self.assertEquals(exit_code, e.code)
        else:
            self.fail("expected test to fail with an error like {!r}".format(re_str))

    def tearDown(self):
        self.test_log.pop_application()

    @classmethod
    def _setup_sim(cls):
        cls.simulator = Simulator()
        cls.simulator.activate()
        cls.simulator.set_version('2.0.0')
        cls.mock_infinibox = InfiniBox(cls.simulator, auth=('infinidat', '123456'))
        cls.mock_infinibox.login()

    @classmethod
    def _setup_pool(cls):
        Pool = cls.mock_infinibox.objects.pools
        cls.pool = Pool.create(name='testpool', virtual_capacity=2 * TiB,
                               physical_capacity=2 * TiB)

    @classmethod
    def _setup_volumes(cls):
        volnames = ["{}_{}".format(cls.VOL_PREFIX, i)
                    for i in range(0, cls.MAXVOLS)]
        for volname in volnames:
            vol = cls.mock_infinibox.objects.volumes.create(pool=cls.pool,
                                                            name=volname,
                                                            size=5 * GiB)
            snapnames = set(name_template.format(volname=volname,
                                                 basename=basename,
                                                 date=creation_date)
                            for name_template in cls.TEST_NAME_TEMPLATES
                            for creation_date in cls.CREATION_DATES
                            for basename in cls.BASENAMES)
            for snapname in snapnames:
                vol.create_snapshot(name=snapname)

    @classmethod
    def setup_class(cls):
        cls._setup_sim()
        cls._setup_pool()
        cls._setup_volumes()

    @classmethod
    def teardown_class(self):
        pass

    # def setup(self):
    #     self.log_handler = logbook.TestHandler()
    #     self.log_handler.push_thread()

    # def teardown(self):
    #     self.log_handler.pop_thread()
    def _create_config_file(self, infinibox_username=u'infinidat', infinibox_password=u'123456'):
        config = {
            'datastores': {'*': u''},
            'nomemory': {'*': u'', '!*xxx*': u''},
            'logging': {'syslog': True, 'rotation': False, 'logpath': u'/var/log/infinisnap/infinisnap.log', 'quiet': False},
            'infinibox': {'username': infinibox_username, 'host': u'ibox515', 'password': infinibox_password},
            'global': {'snap_basename': u'daily', 'snaps_to_keep': '7'},
            'noquiesce': {'*': u'', '!xxx': u''},
            'vcenter': {'username': u'administrator', 'host': u'vsc-vc.lab.il.infinidat.com', 'password': u'vmwareesx'},
            'vms': {'*': u''}
        }
        return config

    def _create_simulator_config_file(self, *args, **kwargs):
        config = self._create_config_file(*args, **kwargs)
        config['infinibox']['host'] = self.simulator
        return config

    def _create_vmfs_datastore(self, name, disknames):
        extent = [Mock(diskName=s) for s in disknames]
        disk = Mock()
        disk.name = name
        disk.summary = Mock(type='VMFS')
        disk.info = Mock(vmfs=Mock(extent=extent))
        return disk

    def _create_nfs_datastore(self, name):
        disk = Mock()
        disk.name = name
        disk.summary = Mock(type='NFS')
        disk.info = Mock()
        return disk

    def _create_datastores(self):
        # datastore.summary.type
        # datastore.info.vmfs.extent
        ds1 = self._create_vmfs_datastore(
            'FirstDatastore', ['naa.6742b0f077777777', 'naa.6742b0f077777778'])
        ds2 = self._create_vmfs_datastore(
            'Second Datastore with space', ['naa.6742b0f077777779'])
        ds3 = self._create_vmfs_datastore(
            'Third non infinidat datastore', ['naa.2020202020202020'])
        ds4 = self._create_nfs_datastore('NFS datastore')

        datastores = [ds1, ds2, ds3, ds4]
        return datastores

    def _create_vm(self, name, datastore, template=False, power=True):
        def create_snapshot(snapshot, name, childSnapshotList, datastore):
            return [Mock(
                    snapshot=snapshot,
                    name=name,
                    datastore=datastore,
                    childSnapshotList=childSnapshotList
                    )]

        snap3 = create_snapshot('vim.vm.Snapshot:snapshot-426', 'test3', [], datastore)
        snap2 = create_snapshot('vim.vm.Snapshot:snapshot-425', 'test2', snap3, datastore)
        snap1 = create_snapshot('vim.vm.Snapshot:snapshot-424', 'test1', snap2, datastore)
        snapshot = Mock(rootSnapshotList=snap1)
        if power:
            powerstate = 'poweredOn'
        else:
            powerstate = 'poweredOff'
        vm = Mock(
            summary=Mock(
                config=Mock(
                    template=template),
                runtime=Mock(
                    powerState=powerstate)),
            name=name,
            datastore=datastore,
            snapshot=snapshot
        )
        return vm

    def _create_vms(self):
        names = ['vm1', 'vm2', 'vm3', 'vm4', 'vm5',
                 'vm6', 'vm7', 'vm8', 'template1', 'vm9-off', 'vm10-off']
        esx_host_names = ['esxhost1', 'esxhost2', 'esxhost3']
        datastores = self._create_datastores()
        esx_hosts = list()
        for host in esx_host_names:
            vms = list()
            for name in names:
                vm = self._create_vm(host + '_' + name, datastores, 'template' in name, 'off' not in name)
                vms.append(vm)
            esx_host = Mock(name=host, vm=vms)
            esx_hosts.append(esx_host)
        return esx_hosts

    def _mock_client(self):
        client = Mock()
        client.get_datastores.return_value = self._create_datastores()
        client.get_host_systems.return_value = self._create_vms()
        return client

    def test_validate_arguments(self):
        config = {}
        conf.update_arguments_to_config({}, config)
        validate, error = conf.validate_config(config)

    def test_validate_arguments_wrong(self):
        config = {}
        conf.update_arguments_to_config({'--retain': '1'}, config)
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_right(self):
        config = self._create_config_file()
        validate, error = conf.validate_config(config)
        assert validate

    def test_validate_arguments_snaps_to_keep_large(self):
        config = self._create_config_file()
        config['global']['snaps_to_keep'] = '300'
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_snaps_to_keep_small(self):
        config = self._create_config_file()
        config['global']['snaps_to_keep'] = '0'
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_snaps_to_keep_not_digit(self):
        config = self._create_config_file()
        config['global']['snaps_to_keep'] = 'notdigit'
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_retain_and_purge(self):
        config = self._create_config_file()
        config['global']['snaps_to_keep'] = '5'
        config['infinibox']['purge'] = True
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_no_retain_no_purge(self):
        config = self._create_config_file()
        del config['global']['snaps_to_keep']
        validate, error = conf.validate_config(config)
        assert not validate

    def test_validate_arguments_missing_section(self):
        config = self._create_config_file()
        del config['vcenter']
        validate, error = conf.validate_config(config)
        assert not validate

    def test_datastores(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        assert len(datastores) == 2

    def test_get_datastores_on_block_storage(self):
        client = self._mock_client()
        all_datastores = client.get_datastores()
        vmfs_datastores = vmware.get_datastores_on_block_storage(
            all_datastores)
        assert len(vmfs_datastores) == 3

    def test_iter_all_vms(self):
        client = self._mock_client()
        vms = list(vmware.iter_all_vms(client))
        assert len(vms) == 24

    def test_list_all_snapshots(self):
        client = self._mock_client()
        vms = list(vmware.iter_all_vms(client))
        # print list(vms), type(vms)
        vmname, vmdatastores, vm = vms[0]
        snapshots = vmware.list_all_snapshots(vm)
        assert len(snapshots) == 3

    def test_get_datastores_to_snapshot_all(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        patterns = {'*': u''}
        result_datastores = conf.get_datastores_to_snapshot(datastores, patterns)
        assert len(result_datastores) == 2

    def test_get_datastores_to_snapshot_specific(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        patterns = {'*second*': u''}
        result_datastores = conf.get_datastores_to_snapshot(datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_exclude(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        patterns = {'*': u'', '!*second*': u''}
        result_datastores = conf.get_datastores_to_snapshot(datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_space(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        patterns = {'Second Datastore*': u''}
        result_datastores = conf.get_datastores_to_snapshot(datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_zero(self):
        client = self._mock_client()
        datastores = vmware.get_all_infinidat_datastores(client)
        patterns = {'blahblah': u''}
        result_datastores = conf.get_datastores_to_snapshot(datastores, patterns)
        assert len(result_datastores) == 0

    def test_connect_to_infinibox(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.connect_to_infinibox(config, ssl=False)
        # ibox.login()
        assert ibox.login()

    def test_infinibox_validate_wrong_password(self):
        config = self._create_simulator_config_file(infinibox_username='blah', infinibox_password='blah')
        os.unlink('/tmp/snaprotator_cookie')
        with self.assert_exits_with_error('username/password error', exit_code=errors.EXIT_CODE_IBOX_LOGIN_FAILED):
            ibox = snaprotator.connect_to_infinibox(config, ssl=False)
            snaprotator.infinibox_validate(ibox)

    def test_infinibox_validate_wrong_state(self):
        config = self._create_config_file()
        config['infinibox']['host'] = Simulator()
        with self.assert_exits_with_error('Cannot connect ', exit_code=errors.EXIT_CODE_IBOX_NETWORK_ERROR):
            ibox = snaprotator.connect_to_infinibox(config, ssl=False)
            snaprotator.infinibox_validate(ibox)

    def test_infinibox_validate_too_many_snapshots(self):
        infinibox.MAX_SYSTEM_SNAPSHOTS = 10
        config = self._create_simulator_config_file()
        ibox = snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        with self.assert_exits_with_error('Maximum number of snapshot',
                                          exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_SYSTEM):
            snaprotator.infinibox_validate(ibox)

    def test_infinibox_validate_wrong_version(self):
        sim = Simulator()
        sim.activate()
        sim.set_version('9.9.0')
        infinibox = InfiniBox(sim, auth=('infinidat', '123456'))
        config = self._create_config_file()
        config['infinibox']['host'] = sim
        ibox = snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        with self.assert_exits_with_error('not supported by this version',
                                          exit_code=errors.EXIT_CODE_IBOX_VERSION_UNSUPPORTED):
            snaprotator.infinibox_validate(ibox)

    def test_volume_validate_nonexisting(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        with self.assert_exits_with_error('Failed to find volume', exit_code=errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND):
            snaprotator.volume_validate(ibox, 'non_existing_volume')

    def test_volume_validate_too_many_snapshots(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        infinibox.MAX_VOL_SNAPSHOTS = 10
        with self.assert_exits_with_error('Too many snapshots',
                                          exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_VOLUME):
            volume, snapshots = snaprotator.volume_validate(ibox, 'testvol_0')


    def test_filter_irrelevant_snapshots(self):
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, 'testvol_0')
        for basename in self.BASENAMES:
            self.assertEqual(len(rotation.filter_irrelevant_snapshots(snapshots, 'testvol_0', basename)),
                             len(self.CREATION_DATES))

    def test_sort_and_filter_snapshots(self):
        volume, snapshots = snaprotator.volume_validate(
            self.mock_infinibox, 'testvol_0')
        for basename in self.BASENAMES:
            sorted_snapshots = rotation.sort_and_filter_snapshots(
                snapshots, 'testvol_0', basename)
            for index, date in enumerate(sorted(self.CREATION_DATES)):
                self.assertRegexpMatches(sorted_snapshots[index].get_name(), re.compile(date))

    def test_trim_number_of_snapshots(self):
        volname = 'testvol_0'
        for basename in self.BASENAMES:
            volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
            rotation.trim_number_of_snapshots(self.mock_infinibox, snapshots, volname, basename, 2, False)
            volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
            sorted_snapshots = rotation.sort_and_filter_snapshots(snapshots, volname, basename)
            self.assertEqual(len(sorted_snapshots), 2)

    def test_trim_number_of_snapshots_retantion_bigger_than_num_of_snaps(self):
        volname = 'testvol_0'
        for basename in self.BASENAMES:
            volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
            rotation.trim_number_of_snapshots(self.mock_infinibox, snapshots, volname, basename, 100, False)
            volume, snapshots_after_trim = snaprotator.volume_validate(self.mock_infinibox, volname)
            self.assertEqual(len(snapshots), len(snapshots_after_trim))

    def test_trim_number_of_snapshots_wrong_vol_and_basename(self):
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, 'testvol_0')
        rotation.trim_number_of_snapshots(self.mock_infinibox, snapshots, 'blah', 'blah', 2, False)
        self.assertFalse(self.test_log.has_info(re.compile('Deleted')), self.test_log.formatted_records)
        self.assertFalse(self.test_log.has_error(re.compile('Failed')), self.test_log.formatted_records)

    def test_trim_number_of_snapshots_cant_delete(self):
        volname, basename = 'testvol_3', 'daily'
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
        test_host = self.mock_infinibox.hosts.create(name='delete_guardian')
        for snap in snapshots:
            test_host.map_volume(snap)
        rotation.trim_number_of_snapshots(self.mock_infinibox, snapshots, volname, basename, 2, False)
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
        sorted_snapshots = rotation.sort_and_filter_snapshots(snapshots, volname, basename)
        self.assertTrue(self.test_log.has_warning(re.compile('Failed')), self.test_log.formatted_records)
        self.assertEqual(len(sorted_snapshots), len(self.CREATION_DATES))

    def test_create_latest_snapshot(self):
        volname, basename = 'testvol_4', 'daily'
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
        sorted_snapshots = rotation.sort_and_filter_snapshots(snapshots, volname, basename)
        rotation.create_latest_snapshot(self.mock_infinibox, volume, basename, False, False)
        volume, snapshots_after_create = snaprotator.volume_validate(self.mock_infinibox, volname)
        sorted_snapshots_after_create = rotation.sort_and_filter_snapshots(snapshots_after_create, volname, basename)
        self.assertEquals(len(sorted_snapshots_after_create), len(sorted_snapshots) + 1)

    def test_create_latest_snapshot_new_basename(self):
        volname, basename = 'testvol_4', 'new_basename'
        volume, snapshots = snaprotator.volume_validate(self.mock_infinibox, volname)
        sorted_snapshots = rotation.sort_and_filter_snapshots(snapshots, volname, basename)
        rotation.create_latest_snapshot(self.mock_infinibox, volume, basename, False, False)
        volume, snapshots_after_create = snaprotator.volume_validate(self.mock_infinibox, volname)
        sorted_snapshots_after_create = rotation.sort_and_filter_snapshots(snapshots_after_create, volname, basename)
        self.assertEquals(len(sorted_snapshots_after_create), len(sorted_snapshots) + 1)

    def test_create_latest_snapshot_creation_failed(self):
        rotation.create_latest_snapshot(self.mock_infinibox, None, None, False, False)
        self.assertTrue(self.test_log.has_warning(re.compile('Failed')), self.test_log.formatted_records)
