import re
import sys
from time import sleep
from logsetup import UserLogger

log = UserLogger(__name__)


POWERTOOLS_PACKAGE_NAME = "com.infinidat.powertools"


def get_datastores_on_block_storage(datastores):
    return [datastore for datastore in datastores if datastore.summary.type == 'VMFS']


def is_infinibox_backed_vmfs_datastore(datastore):
    from functools import partial
    pred = partial(re.match, 'naa.6742b0f0')
    return any(pred(extent.diskName) for extent in datastore.info.vmfs.extent)


def get_all_infinidat_datastores(client):
    all_datastores = client.get_datastores()
    block_datastores = get_datastores_on_block_storage(all_datastores)
    return [datastore for datastore in block_datastores if is_infinibox_backed_vmfs_datastore(datastore)]


def iter_all_vms(client):
    esx_hosts = client.get_host_systems()
    for esx_host in esx_hosts:
        for vm in esx_host.vm:
            # only add vm if not template and is powered on
            if not vm.summary.config.template and vm.summary.runtime.powerState == 'poweredOn':
                yield vm.name, ",".join(datastore.name for datastore in vm.datastore), vm


def get_all_vms_on_datastore(datastore):
    return [vm for vm in datastore.vm]


def get_all_vms_on_datastore_with_single_datastore(datastore):
    return [vm for vm in datastore.vm if len(vm.datastore) == 1]


def find_vmware_power_tools_ip(client):
    from urlparse import urlparse
    try:
        e = [e for e in client.service_content.extensionManager.extensionList if e.key == POWERTOOLS_PACKAGE_NAME][0]
        # http://1.2.3.4/something/somethingelse
        urlparts = urlparse(e.server[0].url)
        # extract ip/name
        return urlparts.netloc
    except:
        return None


def iter_filtered_datastores_block_devices(client, filtered_datastores):
    def get_infinidat_devices_under_vmfs_datastore(datastore, infinidat_devs):
        underlying_devs = set(re.sub('^naa\.6', '', extent.diskName) for extent in datastore.info.vmfs.extent)
        return underlying_devs & infinidat_devs

    from infi.storagemodel.vmware import with_host
    esx_hosts = client.get_host_systems()
    for esx_host in esx_hosts:
        with with_host(client, esx_host):
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
            infinidat_devs = set(re.sub('^naa\.6', '', dev.get_canonical_name())
                                 for dev in get_infinidat_block_devices())
            datastores_to_remove = []
            for datastore in filtered_datastores:
                infinidat_devices = get_infinidat_devices_under_vmfs_datastore(datastore, infinidat_devs)
                if infinidat_devices:
                    datastores_to_remove.append(datastore)
                    yield datastore, infinidat_devices
            for d in datastores_to_remove:
                filtered_datastores.remove(d)
        if not filtered_datastores:
            break


def create_snapshot_for_vm(client, snap_basename, vm, quiesce=True, memory=False):
    log.uinfo('Creating snapshot for vm {} (quiesce = {} memory = {})', vm.name, quiesce, memory)
    task = vm.CreateSnapshot_Task(name=snap_basename, memory=memory, quiesce=quiesce)
    client.wait_for_task(task)
    snapshot = task.info.result
    log.debug('Result snapshot is {}', snapshot)
    return snapshot


def initiate_snapshot_for_vm(client, snap_basename, vm, quiesce=True, memory=False):
    vm_name = "<unknown>"
    try:
        vm_name = vm.name
        log.uinfo('Creating snapshot for vm {} (quiesce = {} memory = {})', vm_name, quiesce, memory)
        task = vm.CreateSnapshot_Task(name=snap_basename, memory=memory, quiesce=quiesce)
    except:
        log.uwarn('Snapshot creation failed for VM {}', vm_name)
        task = None
    return task


def wait_for_result(task):
    while task.info.result is None:
        sleep(0.1)
    return task.info.result


def get_datastore_string_name(datastore):
    return str(datastore)[5:-1]


def get_vm_string_name(vm):
    return str(vm)[5:-1]


def initiate_remove_snapshot(vm_snapshot_pair):
    vm, snapshot = vm_snapshot_pair
    from pyVmomi import vim, vmodl
    task = None
    try:
        log.uinfo('Initiate snapshot removal for VM {} ({})', vm.name, snapshot)
        task = snapshot.RemoveSnapshot_Task(removeChildren=False)
    except vim.ManagedObjectNotFound:
        log.uwarn('The object has already been deleted or has not been completely created')
    except vmodl.fault.HostNotConnected as e:
        log.uwarn("Can't delete VM snapshot. {}", e.msg)
    except:
        log.uwarn("Unknown VM snapshot remove error. {}", str(sys.exc_info()[0]))
    return task


def remove_snapshot(client, snapshot):
    from pyVmomi import vim, vmodl
    try:
        task = snapshot.RemoveSnapshot_Task(removeChildren=False)
        client.wait_for_task(task)
    except vim.ManagedObjectNotFound:
        log.uwarn('The object has already been deleted or has not been completely created')
    except vmodl.fault.HostNotConnected as e:
        log.uwarn("Can't delete VM snapshot. {}", e.msg)
    except:
        log.uwarn("Unknown VM snapshot remove error. {}", str(sys.exc_info()[0]))


def list_all_snapshots(vm):
    def recursive_list_all_snapshots(snap, attr='childSnapshotList'):
        from itertools import chain
        snap_list = getattr(snap, attr, [])
        return snap_list + list(chain(*[recursive_list_all_snapshots(o) for o in snap_list]))
    return recursive_list_all_snapshots(vm.snapshot, 'rootSnapshotList')


def cleanup_infinidat_snapshots(client, snap_basename,  vm):
    snapshots = list(list_all_snapshots(vm))
    try:
        for snap in snapshots:
            log.debug('Snap name: {} Snap obj: {}', snap.name, snap.snapshot)
        for snapshot in snapshots:
            if snapshot.name == snap_basename:
                log.uinfo('Deleting snapshot {} for vm {}', snapshot.name, vm.name)
                remove_snapshot(client, snapshot.snapshot)
    except:
        log.uwarn('Unknown error in VMWare cleanup', exc_info=True)


def get_task_status(task):
    message = None
    if task.info.state == 'error':
        if hasattr(task.info.error, 'reason'):
            message = task.info.error.reason
        elif hasattr(task.info.error, 'msg'):
            message = task.info.error.msg
        elif hasattr(task.info.error, 'message'):
            message = task.info.error.message
        else:
            message = 'Unknown'
    return task.info.state, message
